/**CFile***********************************************************************

  FileName    [bmcRbc2bexp.c]

  PackageName [bmc]

  Synopsis    [Convert an RBC to corresponding Boolean Expression]

  Description []

  SeeAlso     []

  Author      [Lorenzo Delana]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

#include "rbcInt.h" /* for RBCVAR and other definitions */

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
/**Struct**********************************************************************

  Synopsis    [Rbc2bexpDfsData]

  Description [The data structure used for DFS traversal of RBC]

  SeeAlso     []

******************************************************************************/
struct Rbc2bexpDfsData {
  array_t * stack;
  int head;
};

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/* Primitives for stack used in RBC traversal */
static void Rbc2bexpDfsData_push ARGS((struct Rbc2bexpDfsData *, node_ptr));
static node_ptr Rbc2bexpDfsData_head ARGS((struct Rbc2bexpDfsData *));
static node_ptr Rbc2bexpDfsData_pop  ARGS((struct Rbc2bexpDfsData *));

/* RBC traversal functions */
static int  Rbc2bexp_Set   ARGS((rbc_ptr rbc, char * Rbc2bexpData, int sign));
static void Rbc2bexp_First ARGS((rbc_ptr rbc, char * Rbc2bexpData, int sign));
static void Rbc2bexp_Back  ARGS((rbc_ptr rbc, char * Rbc2bexpData, int sign));
static void Rbc2bexp_Last  ARGS((rbc_ptr rbc, char * Rbc2bexpData, int sign));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Given an rbc, constructs the corresponding boolean
                      expression]

  Description        [Descends the structure of the RBC with dag-level 
  primitives.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_Rbc2Bexp (rbc_ptr rbc)
{
  Dag_DfsFunctions_t Rbc2bexpFunctions;
  struct Rbc2bexpDfsData Rbc2bexpData;
  node_ptr ret;

  /* Cleaning the user fields. */
  Dag_Dfs(rbc, &dag_DfsClean, NIL(char));

  /* Setting up the DFS functions. */
  Rbc2bexpFunctions.Set        = Rbc2bexp_Set;
  Rbc2bexpFunctions.FirstVisit = Rbc2bexp_First;
  Rbc2bexpFunctions.BackVisit  = Rbc2bexp_Back;
  Rbc2bexpFunctions.LastVisit  = Rbc2bexp_Last;
 
 /* Setting up the DFS data. */
  /* TODO??: optimizations on the quantity of nodes */
  Rbc2bexpData.stack = array_alloc (node_ptr, 10);
  Rbc2bexpData.head = -1;

  /* Calling DFS on f. */
  Dag_Dfs(rbc, &Rbc2bexpFunctions, (char*)(&Rbc2bexpData));
  ret = Rbc2bexpDfsData_head (&Rbc2bexpData);

  array_free(Rbc2bexpData.stack);

  return ret;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Sets a node into the stack]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void Rbc2bexpDfsData_push (struct Rbc2bexpDfsData * data, 
				  node_ptr value)
{
  (data -> head) ++;

  array_insert(node_ptr, data -> stack, data -> head, value);
}

/**Function********************************************************************

  Synopsis           [Rbc2bexpDfsData_head]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static node_ptr Rbc2bexpDfsData_head (struct Rbc2bexpDfsData * data)
{
  /* there is space available into the stack: */
  nusmv_assert ((data -> head) != (-1));

  return (node_ptr)array_fetch(node_ptr, data->stack, data->head);
}

/**Function********************************************************************

  Synopsis           [Rbc2bexpDfsData_pop]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static node_ptr Rbc2bexpDfsData_pop (struct Rbc2bexpDfsData * data)
{
  node_ptr value = (node_ptr)Rbc2bexpDfsData_head(data);

  (data->head) --;

  return(value);
}

/**Function********************************************************************

  Synopsis           [Rbc2bexpSet]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static int Rbc2bexp_Set (rbc_ptr rbc, char * Rbc2bexpData, int sign)
{
  return (-1);
}

/**Function********************************************************************

  Synopsis           [Rbc2bexpFirst]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void Rbc2bexp_First (rbc_ptr rbc, char * Rbc2bexpData, int sign)
{
  return;
}

/**Function********************************************************************

  Synopsis           [Rbc2bexp_Back]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void Rbc2bexp_Back (rbc_ptr rbc, char * Rbc2bexpData, int sign)
{
  return; 
}

/**Function********************************************************************

  Synopsis           [Rbc2bexp_Last]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static void Rbc2bexp_Last (rbc_ptr rbc, char* Rbc2bexpData, int sign)
{
  int identifier = 0;
  int time, var_id;
  node_ptr left, right;

  switch (rbc -> symbol) {
  case RBCTOP:
    if (sign == RBC_FALSE) {
      Rbc2bexpDfsData_push(
	     (struct Rbc2bexpDfsData *)Rbc2bexpData, Bmc_WffMkFalsity());
    }
    else {
      Rbc2bexpDfsData_push(
	     (struct Rbc2bexpDfsData *)Rbc2bexpData, Bmc_WffMkTruth());
    }
    break;    

  case RBCVAR:
    /* substitute the variable index, in the stack, with its correspondent
       state variable */
    
    time = Bmc_VarIndex2Time((int)(rbc->data));
    var_id = Bmc_VarIndex2StateVarIndex((int)(rbc->data));

    if (sign == RBC_FALSE) {
      Rbc2bexpDfsData_push (
	     (struct Rbc2bexpDfsData *)Rbc2bexpData,
	     Bmc_WffMkNot( Bmc_WffMkXopNext(Bmc_GetIthStatevar(var_id), time) )
		           );
    }
    else {
      Rbc2bexpDfsData_push((struct Rbc2bexpDfsData *)Rbc2bexpData,
			   Bmc_WffMkXopNext(Bmc_GetIthStatevar(var_id), time));
    }
    break;
  case RBCAND:
  case RBCIFF:
    /* get the left bexp from the stack */
    right = Rbc2bexpDfsData_pop ((struct Rbc2bexpDfsData *)Rbc2bexpData);

    /* get the right bexp from the stack */
    left = Rbc2bexpDfsData_pop ((struct Rbc2bexpDfsData *)Rbc2bexpData);

    switch (rbc -> symbol) {
    case RBCAND:
      identifier = AND;
      break;
    case RBCIFF:
      identifier = IFF;
      break;
    }

    if (sign == RBC_FALSE) {
      Rbc2bexpDfsData_push (
		      (struct Rbc2bexpDfsData *)Rbc2bexpData,
		      Bmc_WffMkNot (find_node (identifier, left, right))
		           );
    }
    else {
      Rbc2bexpDfsData_push ( (struct Rbc2bexpDfsData *)Rbc2bexpData,
			     find_node (identifier, left, right) );
    }
    break;
    
  default:
    /* execution should never be here: */
    internal_error("rbc -> symbol had an invalid value: %d\n", rbc -> symbol);
    nusmv_assert(false);
  }
}








